from Products.CMFMetadata.MetadataTool import MetadataTool

def install(self):
    portal = self.portal_url.getPortalObject()
    portal._setObject(MetadataTool.id, MetadataTool())

    return 'CMF Advanced Metadata Tool Installed'



"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
# Zope
from Acquisition import aq_base
# Annotations
from Products.Annotations.AnnotationTool import Annotations
# Formulator
from Products.Formulator import Form
# SilvaMetadata
from Access import invokeAccessHandler
import Configuration
from ZopeImports import *
from Namespace import MetadataNamespace, BindingRunTime
from Binding import ObjectDelegate, encodeElement
from Exceptions import BindingError
from Compatibility import IActionProvider, IPortalMetadata, ActionProviderBase
from Compatibility import getContentType, getContentTypeNames

class MetadataTool(UniqueObject, Folder, ActionProviderBase):

    id = 'portal_metadata'
    meta_type = 'Advanced Metadata Tool'
    title =  meta_type

    manage_options = (
        {'label':'Overview',
         'action':'manage_main'},

        {'label':'Metadata Sets',
         'action':'%s/manage_workspace' % Configuration.MetadataCollection},

        {'label':'Type Mapping',
         'action':'%s/manage_workspace' % Configuration.TypeMapping},
        )

    __implements__ = (IActionProvider, IPortalMetadata)

    _actions = []

    security = ClassSecurityInfo()
    #security.declareProtected('metadata_overview', Configuration.pMetadataManage)
    #metadata_overview = DTMLFile('ui/MetadataToolOverview', globals())
    manage_main = DTMLFile('ui/MetadataToolOverview', globals())

    def __init__(self):
        pass

    #################################
    # Action Provider Interface
    def listActions(self, info=None):
        actions = []
        for set in self.getCollection().getMetadataSets():
            if set.use_action_p is not None and set.action is not None:
                actions.append(set.action)
        return self._actions

    #################################
    # Metadata interface

    ## site wide queries

    # this is the wrong tool to be asking.
    #def getFullName(self, userid):
    #    return userid

    # this is just lame, assumes global publisher for a site
    #def getPublisher(self):
    #    pass

    ## dublin core hardcodes :-(
    # we don't have vocabulary implementation yet.

    def listAllowedSubjects(self, content=None):
        catalog = getToolByName(self, 'portal_catalog')
        return catalog.uniqueValuesFor('Subject')

    def listAllowedFormats(self, content=None):
        catalog = getToolByName(self, 'portal_catalog')
        return catalog.uniqueValuesFor('Format')

    def listAllowedLanguages(self, content=None):
        catalog = getToolByName(self, 'portal_catalog')
        return catalog.uniqueValuesFor('Language')

    def listAllowedRights(self, content=None):
        catalog = getToolByName(self, 'portal_catalog')
        return catalog.uniqueValuesFor('Rights')

    #################################
    ## validation hooks
    def setInitialMetadata(self, content):
        binding = self.getMetadata(content)
        sets = binding.getSetNames()
        # getting the set metadata will cause its
        # initialization if nots already initialized
        for s in sets:
            binding[s]

    def validateMetadata(self, content):
        binding = self.getMetadata(content)
        sets = binding.getSetNames()
        for s in sets:
            data = binding[s]
            binding.validate(data, set_id=s)

    #################################
    ## new interface

    def getCollectionForCategory(self, category=''):
        """return a container containing all known metadata sets
        """
        collections = self._getOb(Configuration.MetadataCollection)
        return [coll for coll in collections if coll.getCategory() == category]
    
    def getCollection(self):
        """return a container containing all known metadata sets
        """
        return self._getOb(Configuration.MetadataCollection)

    def getTypeMapping(self):
        """ return the mapping of content types to metadata sets """
        return self._getOb(Configuration.TypeMapping)

    def getMetadataSet(self, set_id):
        """ get a particular metadata set """
        return self.getCollection().getMetadataSet(set_id)

    def getMetadataSetFor(self, metadata_namespace):
        """ get a particular metadata set by its namespace """
        return self.getCollection().getSetByNamespace(metadata_namespace)

    def getMetadata(self, content):
        """
        return a metadata binding adapter for a particular content
        object. a bind adapter encapsulates both metadata definitions,
        data, and policy behavior into an api for manipulating and
        introspecting metadata
        """
        ct = getContentType(content)

        if not ct in getContentTypeNames(self):
            raise BindingError(
                "invalid content type %s for metadata system" % ct
                )
        return invokeAccessHandler(self, content)

    def getMetadataValue(self, content, set_id, element_id):
        """Get a metadata value right away. This can avoid
        building up the binding over and over while indexing.

        This really goes to the low-level to speed this up to the maximum.
        It's only going to work for Silva, not CMF.
        """
        # XXX how does this interact with security issues?
        set = self.collection.getMetadataSet(set_id)
        element = set.getElement(element_id)
        annotations = getattr(aq_base(content), '_portal_annotations_', None)

        bind_data = None
        if annotations is not None:
            bind_data = annotations[MetadataNamespace].get(BindingRunTime)
        if bind_data is not None:
            delegate = bind_data.get(ObjectDelegate)
            if delegate is not None:
                content = getattr(content,delegate)()
                annotations = getattr(aq_base(content), '_portal_annotations_', None)

        try:
            saved_data = annotations[MetadataNamespace].get(set.metadata_uri)
        except (TypeError, KeyError):
            saved_data = None
            
        #print 'found it for:', repr((saved_data, content))
            
        # if it's saved, we're done
        if saved_data is not None and saved_data.has_key(element_id):
            return saved_data[element_id]
        # if not, check whether we acquire it, if so, we're done
        if element.isAcquireable():
            aqelname = encodeElement(set_id, element_id)
            try:
                #print 'acquiring?', repr(aqelname)
                return getattr(content, aqelname)
            except AttributeError:
                #print 'attr error on', repr(aqelname)
                pass
        # if not acquired, fall back on default
        return element.getDefault(content=content)

    def getMetadataForm(self, context, set_id):
        """Get a complete Formulator form for a metadata set. This helps
        validating user input.
        """        
        set = self.collection.getMetadataSet(set_id)
        fields = [element.field for element in set.getElements()]
        form = Form.BasicForm().__of__(context)
        form.add_fields(fields)
        return form
    
    # Convenience methods
    
    def initializeMetadata(self):
        # initialize the sets if not already initialized
        collection = self.getCollection()
        for set in collection.getMetadataSets():
            if not set.isInitialized():
                set.initialize()
            
    def addTypesMapping(self, types, setids, default=''):
        for type in types:
            self.addTypeMapping(type, setids, default)
            
    def addTypeMapping(self, type, setids, default=''):
        mapping = self.getTypeMapping()
        chain = mapping.getChainFor(type)
        if chain == 'Default':
            chain = ''
        chain = [c.strip() for c in chain.split(',') if c]
        for setid in setids:
            if setid in chain:
                continue
            chain.append(setid)
        tm = {'type': type, 'chain': ','.join(chain)}
        mapping.editMappings(default, (tm, ))

    def removeTypesMapping(self, types, setids):
        for type in types:
            self.removeTypeMapping(type, setids)
        
    def removeTypeMapping(self, type, setids):
        mapping = self.getTypeMapping()
        chain = mapping.getChainFor(type)
        if chain == 'Default' or chain == '':
            return
        chain = [c.strip() for c in chain.split(',') if c]
        for setid in setids:
            if setid in chain:
                chain.remove(setid)
        tm = {'type': type, 'chain': ','.join(chain)}
        default = mapping.getDefaultChain()
        mapping.editMappings(default, (tm, ))
    
    #################################
    # misc

    def manage_afterAdd(self, item, container):
        initializeTool(self)

    def update(self, RESPONSE):
        """ """
        RESPONSE.redirect('manage_workspace')

def initializeTool(tool):

    from Collection import MetadataCollection
    from TypeMapping import TypeMappingContainer

    collection = MetadataCollection(Configuration.MetadataCollection)
    collection.id = Configuration.MetadataCollection
    tool._setObject(Configuration.MetadataCollection, collection)

    mapping = TypeMappingContainer(Configuration.TypeMapping)
    mapping.id = Configuration.TypeMapping
    tool._setObject(Configuration.TypeMapping, mapping)




"""
description: some really ugly default view.. will likely be removed in the
near future in favor of some zpt views

author: kapil thangavelu
"""

from utils import StringBuffer
from ZopeImports import getToolByName

def getForm(binding, set, framed=1, REQUEST=None, errors=None):
    content = binding.content
    elements = set.getElementsFor(content, mode='edit')
    annotations = getToolByName(content, 'portal_annotations')

    if REQUEST is not None and errors is not None:
        data = REQUEST.form.get(set.getId())
    else:
        data = binding._getData(set.getId(), acquire=0)

    out = StringBuffer()

    if framed:
        print >> out, '<form method="POST" action="edit_metadata"' \
                      ' enctype="multipart/form-data">'
        print >> out, '<input type="hidden" name="mset_id" value="%s">' \
                      % set.getId()

    print >> out, '<h2 class="metadata_header">%s</h2>' % set.getTitle()

    if errors is not None:
        print >> out, '<table class="form_errors">'
        for k,v in errors.items():
            print >> out, '<tr><td>%s: %s</td</tr>' % (k,v)
        print >> out, '</table>'

    print >> out, '<table class="metadata_form">'
    for e in elements:
        print >> out, "<tr><td><b>%s</b></td>" % e.title()
        print >> out, "<tr><td>%s</td>" \
                 % e.renderEdit(data.get(e.getId(), None))

    if framed:
        print >> out, '''<tr><td colspan="2">
        <input type="submit" value="edit %s">
        </td></tr>''' % set.getTitle()

    print >> out, '</table>'

    if framed:
        print >> out, '</form>'

    return out.getvalue()

def getView(binding, set, framed=1):
    content = binding.content
    elements = set.getElementsFor(content, mode='view')
    annotations = getToolByName(content, 'portal_annotations')
    data = annotations.getAnnotations(content, set.metadata_uri)
    out = StringBuffer()

    print >> out, '<table class="metadata_view">'

    for e in elements:
        print >> out, "<tr><td><b>%s</b></td>" % e.title()
        print >> out, "<tr><td>%s</td></td>" \
                      % renderField(e, e.getId(), data.get(e.getId(), None))

    print >> out, '</table>'
    return out.getvalue()

def renderField(element, key, value):
    # XXX formulator specific
    view_method = getattr(element.field.widget, 'render_view', None)
    if view_method is not None:
        return view_method(element.field, key, value)
    elif not value: #shizer
        return ''
    else:
        return str(value)



"""
author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

class NoContext(Exception): pass
class NamespaceConflict(Exception): pass
class ConfigurationError(Exception): pass
class NotFound(AttributeError): pass
class CompatibilityException(Exception): pass
class ImportError(Exception): pass
class ValidationError(Exception): pass
class BindingError(Exception): pass
class XMLMarshallError(Exception): pass

# For use in python scripts
from Products.PythonScripts.Utility import allow_class
allow_class(BindingError)



"""
functions for installing and removing indexes for
metadata elements in a zcatalog.

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from Products.ProxyIndex import ProxyIndex
from Compatibility import index_expression_template, getToolByName
from UserDict import UserDict

def createIndexes(catalog, elements):

    all_indexes = catalog.indexes()

    for e in elements:
        idx_id  = createIndexId(e)
        extra   = createIndexArguements(e)

        if idx_id in all_indexes:
            continue

        catalog.addIndex(idx_id, ProxyIndex.ProxyIndex.meta_type, extra)
        all_indexes.append(idx_id)

    return None

def removeIndexes(catalog, elements):

    all_indexes = catalog.indexes()

    for e in elements:
        idx_id = createIndexId(e)

        if not idx_id in all_indexes:
            continue

        catalog.delIndex(idx_id)
        all_indexes.remove(idx_id)

    return None

def getIndexNamesFor(elements):
    res = []
    for e in elements:
        res.append(createIndexId(e))
    return res

def createIndexId(element):
    ms = element.getMetadataSet()
    return "%s%s" % (ms.metadata_prefix, element.getId())

def createIndexArguements(element):

    d = ProxyIndex.RecordStyle()

    # try to get the element's index construction key/value pair
    if element.index_constructor_args is not None:
        d.update(element.index_constructor_args)

    d['idx_type'] = element.index_type
    d['value_expr'] = createIndexExpression(element)

    # we setup the idx context manually ourselves..
    # proxyindex needs to find the zcatalog in the containement
    # hierarchy to introspect the pluggable indexes.
    d['idx_context'] = getToolByName(element, 'portal_catalog')

    return d

def createIndexExpression(element):
    return index_expression_template % (
        element.getMetadataSet().getId(),
        element.getId()
        )




"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

#################################
### Silva / CMF Compatiblity Flag
# if using Silva set to 0
UsingCMF = 0

#################################
# this module gets populated with
# permissions for these by the compatiblity
# module.
pMetadataView = None
pMetadataEdit = None
pMetadataManage = None

#################################
### internal names
MetadataCollection = 'collection'
TypeMapping = 'ct_mapping'



"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

import Configuration
import Compatibility

import Collection
import Set
import Element
import Binding
import MetadataTool

from Globals import ImageFile
from Products.Annotations.helpers import setupIcon

# Allow Errors to be imported TTW
from Products.PythonScripts.Utility import allow_module
allow_module('Products.SilvaMetadata.Exceptions')

misc_ = {
    'up'     : ImageFile('www/up.gif', globals()),
    'down'   : ImageFile('www/down.gif', globals()),
    'top'    : ImageFile('www/top.gif', globals()),
    'bottom' : ImageFile('www/bottom.gif', globals()),
    # XXX ugh! misc_ here overrides whatever is done in setupIcon apparently,
    # so let's set up metadata icon here again..
    'silva_metadata.png': ImageFile('www/silva_metadata.png', globals()),
    }

setupIcon(MetadataTool.MetadataTool,
          'www/silva_metadata.png', 'SilvaMetadata', globals())



"""
Misc. Utility Functions and classes

Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

class StringBuffer:

    def __init__(self):
        self.buf =[]
    def write(self, data):
        self.buf.append(data)
    def writeln(self, data):
        self.buf.append(data)
        self.buf.append('\n')
    def getvalue(self):
        return ''.join(self.buf)

def make_lookup(seq):
    d = {}
    for s in seq:
        d[s]=None
    return d.has_key

def normalize_kv_pairs(mapping):
    res = {}

    keys = [k for k in mapping.keys() if k.startswith('key')]
    keys.sort()
    vals = [v for v in mapping.keys() if v.startswith('value')]
    vals.sort()

    pairs = zip(keys, vals)

    for k,v in pairs:
        res[mapping[k]]=mapping[v]

    return res







"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
# Python
import types, copy
from UserDict import UserDict
# Zope
from Acquisition import Implicit, aq_base, aq_parent
from Products.ZCatalog.ZCatalog import ZCatalog
# Formulator
from Products.Formulator.Errors import FormValidationError
#
from zExceptions import Unauthorized
from Exceptions import NotFound, BindingError
from Export import ObjectMetadataExporter
from Index import getIndexNamesFor
import Initialize as BindingInitialize
from Namespace import MetadataNamespace, BindingRunTime
from ZopeImports import Interface, ClassSecurityInfo, InitializeClass
from ZopeImports import PersistentMapping, getToolByName

#################################
### runtime bind data keys
AcquireRuntime = 'acquire_runtime'
ObjectDelegate = 'object_delegate'
MutationTrigger = 'mutation_trigger'

#################################
### Acquired Metadata Prefix Encoding
MetadataAqPrefix = 'metadataAq'
MetadataAqVarPrefix = '_VarName_'

_marker = []


class Data(UserDict):
    """
    We use this as to escape some vagaries with the zope security policy
    when using the __getitem__ interface of the binding
    """
    __roles__ = None
    __allow_access_to_unprotected_subobjects__ = 1


class MetadataBindAdapter(Implicit):

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    def __init__(self, content, sets):
        self.content = content
        self.collection = {}
        self.setnames = []
        self.category_to_setnames = {}
        self.cached_values = {}
        
        for set in sets:
            setid = set.getId()
            self.setnames.append(setid)
            self.collection[setid] = set
            category = set.getCategory()
            setnames = self.category_to_setnames.setdefault(category, [])
            setnames.append(setid)
            
    #################################
    ### Views
    security.declarePublic('renderXML')
    def renderXML(self, set_id=None, namespace_key=None):
        """
        return an xml serialization of the object's metadata
        """

        if set_id or namespace_key:
            sets = [self._getSet(set_id, namespace_key)]
        else:
            sets = self.collection.values()

        exporter = ObjectMetadataExporter(self, sets)
        return exporter()

    security.declarePublic('renderElementView')
    def renderElementView(self, set_id, element_id):
        element = self.getElement(set_id, element_id)
        data = self._getData(set_id)
        value = data.get(element_id, None)
        return element.renderView(value)

    security.declarePublic('renderElementEdit')
    def renderElementEdit(self, set_id, element_id):
        element = self.getElement(set_id, element_id)
        data = self._getData(set_id, acquire=0)
        return element.renderEdit(data.get(element_id, None))

    #################################
    ### Validation

    security.declarePublic('validate')
    def validate(self, set_id, data, errors=None):
        """
        validate the data. implicit transforms may be preformed.
        """
        return validateData(self, self.collection[set_id], data, errors)

    security.declarePublic('validateFromRequest')
    def validateFromRequest(self, set_id, REQUEST, errors=None):
        """
        validate from request
        """
        data = REQUEST.form.get(set_id)

        if data is None:
            raise NotFound("Metadata for %s/%s not found" % (
                str(set_id), str(namespace_key)))

        return self.validate(set_id, data.copy(), errors)

    #################################
    ### Storage (invokes validation)

    security.declarePublic('setValues')
    def setValues(self, set_id, data, reindex=0):
        """
        data should be a mutable dictionary

        returns a dictionary of errors if any, or none otherwise
        """
        errors = {}
        data = self.validate(set_id, data, errors)

        if errors:
            return errors

        set = self.collection[set_id]
        self._setData(data, set_id=set_id, reindex=reindex)
        return None

    security.declarePublic('setValuesFromRequest')
    def setValuesFromRequest(self, request, reindex=0):
        """Returns a dictionary of errors if any
        """
        all_errors = {}
        ms = self.service_metadata
        context = self._getAnnotatableObject()
        setnames = self.getSetNames()
        for setname in setnames:
            if setname not in request.form:
                continue
            try:
                form = ms.getMetadataForm(context, setname)
                result = form.validate_all(request.form[setname])
                
                # Remove keys from the result that are supposed to be
                # read-only only
                set = self.getSet(setname)
                elements = set.getElements()
                for element in elements:
                    if not element.isEditable(context):
                        try:
                            del result[element.id]
                        except KeyError, e:
                            pass
                
            except FormValidationError, e:
                all_errors[setname] = errors = {}
                for error in e.errors:
                    errors[error.field_id] = error.error_text
            else:
                self._setData(result, setname)
        return all_errors

    #################################
    ### Discovery Introspection // Definition Accessor Interface

    security.declarePublic('getSetNames')
    def getSetNames(self, category=None):
        """Return the ids of the metadata sets available for this 
        content type.
        """
        if category is None:
            return tuple(self.setnames)
        setnames = self.category_to_setnames.get(category, [])
        return tuple(setnames)

    security.declarePublic('keys')
    keys = getSetNames

    security.declarePublic('getElementNames')
    def getElementNames(self, set_id, mode=None):
        """Given a set identifier return the ids of the elements
        within the set.
        """
        # XXX
        # if mode is not specified this returns all elements of a set.
        # not all elements visible will be viewable/editable
        set = self.collection[set_id]

        if mode is not None:
            elements = set.getElementsFor(self.content, mode=mode)
        else:
            elements = set.getElements()

        return [e.getId() for e in elements]

    security.declarePublic('getSetNameByURI')
    def getSetNameByURI(self, uri):
        for set in self.collection.values():
            if set.metadata_uri == uri:
                return set.getId()
        raise NotFound(uri)

    security.declarePublic('getSet')
    def getSet(self, set_id):
        """To invoke methods on the set requires permissions on
        the set not the content, whereas binding methods
        merely require permissions on the content.
        """
        return self.collection[set_id]

    security.declarePublic('getElement')
    def getElement(self, set_id, element_id):
        return self.getSet(set_id).getElement(element_id)
    
    security.declarePublic('isViewable')
    def isViewable(self, set_id, element_id):
        """
        is the element viewable for the content object
        """
        element = self.collection[set_id].getElement(element_id)
        ob = self._getAnnotatableObject()
        return element.isViewable(ob)

    security.declarePublic('isEditable')
    def isEditable(self, set_id, element_id):
        """
        is the element editable for the content object
        """
        element = self.collection[set_id].getElement(element_id)
        ob = self._getAnnotatableObject()
        return element.isEditable(ob)


    security.declarePublic('listAcquired')
    def listAcquired(self):
        """
        compute and return a list of (set_id, element_id)
        values for all metadata which this binding/content
        acquires from above in the containment hiearchy.
        """
        res = []
        ob = self._getAnnotatableObject()

        for s in self.collection.values():
            sid = s.getId()
            data = self._getData(set_id = sid, acquire=0)
            for e in [e for e in s.getElements() if e.isAcquireable()]:
                eid = e.getId()
                if data.has_key(eid) and data[eid]:
                    continue
                name = encodeElement(sid, e.getId())
                try:
                    value = getattr(ob, name)
                except AttributeError:
                    continue
                # filter out any empty metadata fields
                # defined on ourselves to acquire
                if not hasattr(aq_base(ob), name):
                    res.append((sid, eid))

        return res

    #################################
    ### Data Accessor Interface

    security.declarePublic('get')
    def get(self, set_id, element_id=None, acquire=1, no_defaults=0):
        """
        if element_id is specified, only the value of that element is
        returned, otherwise a Data object is returned. Data objects
        implement a mapping interface.

        the acquire flag determines whether or not metadata acquisition
        will be used in retrieving values, acquired values will override
        default values but not values stored on the object.

        the no_defaults flag specifies whether an element's default values
        should be used. default values are only used when there is no
        value stored on the object.

        the use case for no_defaults is whe using tales defaults to defer
        an element's value to another element within the same set. also when
        no_defaults is used not all elements of the set will nesc. be
        in the data object returned only those which were findable.
        """
        data = self._getData(set_id=set_id,
                             acquire=acquire,
                             no_defaults=no_defaults)
        if element_id is not None:
            return data.get(element_id)
        return data

    def __getitem__(self, key):
        if self.collection.has_key(key):
            return self._getData(key)
        raise KeyError(str(key))

    #################################
    ### RunTime Binding Methods

    security.declarePublic('setObjectDelegator')
    def setObjectDelegator(self, method_name):
        """
        we get and set all metadata on a delegated object,
        method should be a callable method on the object
        (acquiring the method is ok) that should take zero
        args, and return an object. if it doesn't return
        an object, we return the default metadata values
        associated (not a good idea).
        """
        assert getattr(self.content, method_name), \
                       "invalid object delegate %s" % method_name

        bind_data = self._getBindData()
        bind_data[ObjectDelegate]=method_name

    security.declarePrivate('getObjectDelegator')
    def getObjectDelegator(self):
        return self._getBindData().get(ObjectDelegate)

    security.declarePublic('clearObjectDelegator')
    def clearObjectDelegator(self):
        bind_data = self._getBindData()
        try:
            del bind_data[ObjectDelegate]
        except KeyError:
            pass
        # invalidate cache
        self.cached_values = {}

        return None

    security.declarePublic('setMutationTrigger')
    def setMutationTrigger(self, set_id, element_id, method_name):
        """
        support for simple events, based on acquired method
        invocation. major use case.. cache invalidation on
        metadata setting.
        """
        assert getattr(self._getAnnotatableObject(), method_name), \
                       "invalid mutation trigger %s" % method_name

        bind_data = self._getBindData()
        tr = bind_data.setdefault(MutationTrigger, {}).setdefault(set_id, {})
        tr[element_id]=method_name

    security.declarePublic('clearMutationTrigger')
    def clearMutationTrigger(self, set_id, element_id=None):
        """
        clear mutation triggers for a particular set or element.

        if element_id is not specified, clear triggers for
        the entire set.
        """

        bind_data = self._getBindData()
        triggers = bind_data[MutationTrigger]

        if element_id is None:
            try:
                del triggers[set_id]
            except KeyError:
                pass
        else:
            try:
                del triggers[set_id][element_id]
            except KeyError:
                pass
        return None

    #################################
    ### Private

    def _getSet(self, set_id=None, namespace_key=None):
        if set_id:
            return self.collection[set_id]
        elif namespace_key:
            return self._getSetByKey(namespace_key)
        else:
            raise NotFound("metadata set not found %s %s"
                           % (set_id, namespace_key))

    def _getBindData(self):
        annotations = getToolByName(self.content, 'portal_annotations')
        metadata = annotations.getAnnotations(self.content, MetadataNamespace)
        bind_data = metadata.get(BindingRunTime)

        if bind_data is None:
            init_handler = BindingInitialize.getHandler(self.content)
            bind_data = metadata.setdefault(BindingRunTime, PersistentMapping())
            if init_handler is not None:
                init_handler(self.content, bind_data)

        return bind_data

    def _getMutationTriggers(self, set_id):
        bind_data = self._getBindData()
        return bind_data.get(MutationTrigger, {}).get(set_id, [])

    def _getAnnotatableObject(self):
        # check for object delegation
        bind_data = self._getBindData()
        object_delegate = bind_data.get(ObjectDelegate)

        # we want to use the content in its original acquisiton
        # context, but because we retrieve it as an attribute
        # it gets wrapped.. content.__of__(binding).__of__(content)
        # so we remove the outer two wrappers to regain the original
        # context
        content = aq_parent(aq_parent(self.content))

        if object_delegate is not None:
            od = getattr(self.content, object_delegate)
            ob = od()
        else:
            ob = content

        return ob

    def _getData(self, set_id=None, namespace_key=None,
                 acquire=1, no_defaults=0):
        """
        find the metadata for the given content object,
        performs runtime binding work as well.

        """

        set = self._getSet(set_id, namespace_key)

        # cache lookup
        data = self.cached_values.get((acquire, set.getId()))
        if data is not None:
            return data

        using_defaults = []
        ob = self._getAnnotatableObject()

        # get the annotation data
        annotations = getToolByName(ob, 'portal_annotations')
        metadata = annotations.getAnnotations(ob, MetadataNamespace)

        saved_data = metadata.get(set.metadata_uri)
        data = Data()

        sid = set.getId()
        element_ids = self.getElementNames(sid)

        if saved_data is None and no_defaults:
            pass
        elif saved_data is None:
            # use the sets defaults
            defaultvalues = set.getDefaults(content=ob)
            data.update(defaultvalues)
            # record which elements we used default values for
            using_defaults = element_ids
        else:
            # make a copy so we can modify with acq metadata
            data.update(saved_data)

            if not no_defaults:
                # update individual elements with default values
                # if they don't have a saved value.
                for eid in element_ids:
                    if data.has_key(eid):
                        continue
                    defaultvalue = set.getElement(eid).getDefault(content=ob)
                    data[eid] = defaultvalue
                    using_defaults.append(eid)

        # cache metadata
        self.cached_values[ (acquire, set_id) ]=data

        if not acquire:
            return data

        # get the acquired metadata
        hk = data.has_key
        for e in [e for e in set.getElements() if e.isAcquireable()]:
            eid = e.getId()
            if hk(eid) and data[eid] and not eid in using_defaults:
                continue
            aqelname = encodeElement(sid, eid)
            try:
                val = getattr(ob, aqelname)
            except AttributeError:
                continue
            data[eid]=val

        return data

    def _setData(self, data, set_id=None, namespace_key=None, reindex=0):

        set = self._getSet(set_id, namespace_key)
        set_id = set.getId()

        # check for delegates
        ob = self._getAnnotatableObject()

        # filter based on write guard and whether field is readonly
        all_elements = set.getElements()
        all_eids = [e.getId() for e in all_elements]
        elements = [e for e in set.getElementsFor(ob, mode='edit')]
        eids = [e.getId() for e in elements]

        keys = data.keys()

        for k in keys:
            if k in eids:
                continue
            elif k in all_eids:
                raise Unauthorized('Not Allowed to Edit %s in this context' % k)
            else:
                del data[k]

        # fire mutation triggers
        triggers = self._getMutationTriggers(set_id)

        if triggers:
            for k in keys:
                if triggers.has_key(k):
                    try:
                        getattr(ob, triggers[k])()
                    except: # gulp
                        pass

        # update acquireable metadata
        update_list = [e.getId() for e in set.getElements() \
                                 if  e.isAcquireable() and e.getId() in keys]
        sid = set.getId()

        for eid in update_list:
            aqelname = encodeElement(sid, eid)
            value = data[eid]
            if value:
                setattr(ob, aqelname, value)
            else:
                # Try and get rid of encoded attribute on the
                # annotatable object; this will get acquisition
                # of the value working again.
                try:
                    delattr(ob, aqelname)
                except (KeyError, AttributeError), err:
                    pass

        # save in annotations
        annotations = getToolByName(ob, 'portal_annotations')
        metadata = annotations.getAnnotations(ob, MetadataNamespace)

        if metadata.has_key(set.metadata_uri):
            metadata[set.metadata_uri].update(data)
        else:
            metadata[set.metadata_uri] = PersistentMapping(data)

        # invalidate the cache version of the set if any
        # we do a check for cached acquired/non-acquired
        if self.cached_values.has_key((0, set_id)):
            del self.cached_values[(0, set_id)]
        if self.cached_values.has_key((1, set_id)):
            del self.cached_values[(1, set_id)]

        # mark both the content and the annotatable object as changed so
        # on txn commit bindings in other objectspaces get invalidated as well
        ob._p_changed = 1
        self.content._p_changed = 1

        # reindex object
        if reindex:
            reindex_elements = [
                e for e in elements
                if (e.getId() in keys) and e.index_p]
            idx_names = getIndexNamesFor(reindex_elements)
            catalog = getToolByName(ob, 'portal_catalog')
            # cmf compatibility hack
            ZCatalog.catalog_object(catalog, ob, idxs=idx_names)

    def _getSetByKey(self, namespace_key):
        for s in self.collection.values():
            if s.metadata_uri == namespace_key:
                return s
        raise NotFound(str(namespace_key))

InitializeClass(MetadataBindAdapter)

def validateData(binding, set, data, errors_dict=None):
    # XXX completely formulator specific
    from Products.Formulator.Errors import ValidationError

    # Filter out elements not in the data dict, provided the element is
    # not required or the binding already has a value for this element.
    for e in set.getElements():
        eid = e.getId()
        has_a_value = not not binding.get(set.getId(), eid, acquire=0)
        is_required = e.isRequired()

        if hasattr(aq_base(e.field), 'sub_form'):
            # XXX this is really a datetime hack..
            # Fields with subforms will/might have only their marshalled
            # subform ids stored in the data dict. There really isn't a
            # good way to discover which fields are sub form providers,
            # so we just try to introspect.
            # Get one of the subform field ids, just try one.. unfortunately
            # the presence of a subform has little todo with the fields
            # request encoding. sigh.
            sfid = e.field.sub_form.get_field_ids()[0]
            sfkey = e.field.generate_subfield_key(sfid, validation=1)

            if not data.has_key(sfkey) and (not is_required or has_a_value):
                continue

        elif not data.has_key(eid) and (not is_required or has_a_value):
            continue

        try:
            data[eid] = e.validate(data)
        except ValidationError, exception:
            if errors_dict is not None:
                errors_dict[eid] = exception.error_text
            else:
                raise
    return data

def encodeElement(set_id, element_id):
    """
    after experimenting with various mechanisms for doing
    containment based metadata acquisition, using extension class
    acquisition was found to be the quickest way to do the containment
    lookup. as attributes are stored as opaque objects, the current
    implementation decorates the object heirarchy with encoded variables
    corresponding to the metadata specified as acquired. the encoding
    is used to minimize namespace pollution. acquired metadata is only
    specified in this manner on the source object.
    """
    return MetadataAqPrefix + set_id + MetadataAqVarPrefix + element_id

def decodeVariable(name):
    """ decode an encoded variable name... not used """
    assert name.startswith(MetadataAqPrefix)

    set_id = name[len(MetadataAqPrefix):name.find(MetadataAqVarPrefix)]
    e_id = name[name.find(MetadataAqVarPrefix)+len(MetadataAqVarPrefix):]

    return set_id, e_id




"""
Maps Metadata Sets onto Content Types

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from ZopeImports import *
from Compatibility import getContentTypeNames, getToolByName
from Exceptions import ConfigurationError

DEFAULT_MAPPING_NAME = 'Default'

class TypeMappingContainer(Folder):

    meta_type = 'Type Mapping Container'

    manage_options = (

        {'label':'Content Types',
         'action':'manage_main'},

        {'label':'Metadata Tool',
         'action':'../manage_workspace'},
        )

    manage_main = DTMLFile('ui/TypeMappingContainerForm', globals())

    def __init__(self, id):
        self.id = id
        self.default_chain = ''

    def setDefaultChain(self, chain, RESPONSE=None):

        if not verifyChain(self, chain):
            raise ConfigurationError("invalid metadata set")

        self.default_chain = chain

        if RESPONSE is not None:
            return RESPONSE.redirect('manage_workspace')

    def getDefaultChain(self):
        return self.default_chain

    def getChainFor(self, content_type):
        try:
            ctm = self._getOb(content_type)
        except AttributeError:
            return DEFAULT_MAPPING_NAME
        return ctm.getMetadataChain()

    def getMetadataSetsFor(self, content_type):
        try:
            ctm = self._getOb(content_type) # will throw attr error
        except AttributeError:
            return getMetadataSets(self, self.default_chain)
        return ctm.getMetadataSets()

    def getTypeMappings(self):
        return self.objectValues(TypeMapping.meta_type)

    def getTypeMappingFor(self, content_type):
        try:
            ctm = self._getOb(content_type)
        except AttributeError:
            return None
        return ctm

    def editMappings(self, default_chain, type_chains, RESPONSE=None):
        """ """
        self.setDefaultChain(default_chain)

        for tcr in type_chains:
            t = tcr['type']
            tc = tcr['chain']
            tc = tc.strip()
            if tc == DEFAULT_MAPPING_NAME:
                if self.getChainFor(t) != DEFAULT_MAPPING_NAME:
                    self._delObject(t)
            elif tc == self.getChainFor(t):
                continue
            else:
                tm = self.getTypeMappingFor(t)
                if tm is None:
                    self._setObject(t, TypeMapping(t))
                    ctm = self._getOb(t)
                    ctm.setMetadataChain(tc)
                else:
                    tm.setMetadataChain(tc)

        if RESPONSE is not None:
            return RESPONSE.redirect('manage_workspace')

    def getContentTypes(self):
        return getContentTypeNames(self)

class TypeMapping(Folder):

    meta_type = 'Metadata Type Mapping'

    def __init__(self, id):
        self.id = id
        self.chain = None

    def getMetadataChain(self):
        return self.chain

    def setMetadataChain(self, chain):
        if not verifyChain(self, chain):
            raise ConfigurationError("invalid metadata set")
        self.chain = chain

    def getMetadataSets(self):
        return getMetadataSets(self, self.chain)

def getMetadataSets(ctx, chain):
    sets = filter(None, [c.strip() for c in chain.split(',')])
    tool = getToolByName(ctx, 'portal_metadata')
    return map(tool.getMetadataSet, sets)

def verifyChain(ctx, chain):
    try:
        getMetadataSets(ctx, chain)
    except AttributeError:
        return 0
    return 1



"""
Tests for the SilvaMetada.

$Id: test_Metadata.py,v 1.20 2005/09/01 10:28:43 guido Exp $
"""

from unittest import TestSuite, makeSuite, main

from cStringIO import StringIO

from Testing import ZopeTestCase

from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from Products.Annotations.AnnotationTool import AnnotationTool

from Products.Formulator import StandardFields
from Products.Formulator.TALESField import TALESMethod
from Products.SilvaMetadata.MetadataTool import MetadataTool
from Products.SilvaMetadata.Compatibility import getToolByName
from Products.SilvaMetadata.Configuration import UsingCMF

if UsingCMF:
    from Products.CMFCore.CatalogTool import CatalogTool
    from Products.CMFCore.PortalFolder import PortalFolder
    from Products.CMFCore.tests.base.testcase import SecurityTest
    from Products.CMFCore.TypesTool import FactoryTypeInformation as FTI
    from Products.CMFCore.TypesTool import TypesTool
    from Products.CMFCore.utils import getToolByName

    ZopeTestCase.installProduct('Annotations')
    ZopeTestCase.installProduct('ProxyIndex')
    ZopeTestCase.installProduct('Formulator')
    ZopeTestCase.installProduct('CMFCore')
    TestBaseClass = ZopeTestCase.ZopeTestCase
else:
    from Products.Silva.tests import SilvaTestCase
    TestBaseClass = SilvaTestCase.SilvaTestCase

SET_ID = 'ut_md'


def setupFakePermissions(root):
    # copy & paste from ZopeTestCase :-/
    uf = root.acl_users
    uf.userFolderAddUser( '_test_user', 'secret', ['Manager'], [])
    user = uf.getUserById('_test_user').__of__(uf)
    newSecurityManager(None, user)

def setupCMFTools(root):
    root._setObject('portal_types', TypesTool())
    root._setObject('portal_annotations', AnnotationTool())
    root._setObject('portal_metadata', MetadataTool())
    root._setObject('portal_catalog', CatalogTool())

def setupContentTypes(context):
    types_tool = getToolByName(context, 'portal_types')
    types_tool._setObject('Folder', FTI(id='Folder',
                                        title='Folder or Directory',
                                        meta_type=PortalFolder.meta_type,
                                        product='CMFCore',
                                        factory='manage_addPortalFolder',
                                        filter_content_types=0)
                             )

def setupContentTreeCMF(container):
    ttool = getToolByName(container, 'portal_types')
    ttool.constructContent('Folder', container, 'zoo')

    zoo = container._getOb('zoo')

    ttool.constructContent('Folder', zoo, 'mammals')
    ttool.constructContent('Folder', zoo, 'reptiles')

    mammals = zoo._getOb('mammals')
    reptiles = zoo._getOb('reptiles')

    return zoo

def setupContentTreeSilva(self, container):

    zoo = self.add_folder(container, 'zoo', "Zoo Folder")
    self.add_folder(zoo, 'mammals', "Zoo Mammals")
    zoo = self.add_folder(zoo, 'reptiles', "Zoo Reptiles")
    return zoo

def setupCatalog(context):
    catalog = getToolByName(context, 'portal_catalog')
    pass

def setupMetadataSet(context):
    mtool = getToolByName(context, 'portal_metadata')
    collection = mtool.getCollection()
    collection.addMetadataSet(SET_ID,
                              'tmd',
                              'http://www.example.com/xml/test_md')
    set = collection.getMetadataSet(SET_ID)
    set.addMetadataElement('Title',
                           StandardFields.StringField.meta_type,
                           'KeywordIndex',
                           1
                           )
    set.addMetadataElement('Description',
                           StandardFields.StringField.meta_type,
                           'KeywordIndex',
                           1
                           )
    element = set.getElement('Description')
    element.field._edit({'required':0})
    element.editElementPolicy(acquire_p = 1)
    return set

def setupExtendedMetadataSet(context):
    # add some additional metadata fields
    pm = getToolByName(context, 'portal_metadata')
    collection = pm.getCollection()
    set = collection.getMetadataSet(SET_ID)
    set.addMetadataElement('ModificationDate',
                           StandardFields.DateTimeField.meta_type,
                           'DateIndex',
                           1
                           )
    element = set.getElement('ModificationDate')
    element.field._edit_tales({'default':
                        TALESMethod('content/bobobase_modification_time')})
    element.field._edit({'required':0})
    set.addMetadataElement('Languages',
                           StandardFields.LinesField.meta_type,
                           'KeywordIndex',
                           1
                           )
    element = set.getElement('Languages')
    element.field._edit({'required':0})
    ######
    set.initialize()

def setupMetadataMapping(context):
    mtool = getToolByName(context, 'portal_metadata')
    mapping = mtool.getTypeMapping()
    mapping.setDefaultChain('ut_md')
    if not UsingCMF:
        mtool.addTypesMapping(
            ('Silva Folder', ), ('ut_md', 'silva-extra'))


class MetadataTests(TestBaseClass):

    def afterSetUp(self):
        if UsingCMF:
            self.root = self.app
            setupFakePermissions(self.root)
            setupCMFTools(self.root)
            setupCatalog(self.root)
            setupContentTypes(self.root)

        setupMetadataSet(self.root)
        setupMetadataMapping(self.root)
        if UsingCMF:
            setupContentTreeCMF(self.root)
        else:
            setupContentTreeSilva(self, self.root)


class TestSetImportExport(MetadataTests):

    def testImportExport(self):
        pm = getToolByName(self.root, 'portal_metadata')
        collection = pm.getCollection()
        set = collection.getMetadataSet(SET_ID)
        xml = set.exportXML().encode('ascii')
        xmlio = StringIO(xml)
        collection.manage_delObjects([SET_ID])
        collection.importSet(xmlio)
        set = collection.getMetadataSet(SET_ID)
        xml2 = set.exportXML()
        assert xml == xml2, "Import/Export disjoint"


class TestObjectImportExport(MetadataTests):

    def testImportExport(self):
        from Products.ParsedXML.ParsedXML import createDOMDocument
        from Products.SilvaMetadata.Import import import_metadata

        pm = getToolByName(self.root, 'portal_metadata')
        setupExtendedMetadataSet(self.root)
        zoo = self.root.zoo
        mammals = zoo.mammals
        binding = pm.getMetadata(zoo)
        values = binding.get(SET_ID)
        lines = """
        english
        hebrew
        swahili
        urdu
        """
        values.update(
            {'Title':'hello world',
             'Description':'cruel place',
             'Languages':lines }
            )
        binding.setValues(SET_ID, values)
        xml = "<folder>%s</folder>" % binding.renderXML(SET_ID)
        dom = createDOMDocument(xml)
        import_metadata(mammals, dom.childNodes[0])
        mammals_binding = pm.getMetadata(mammals)
        mammal_values = binding.get(SET_ID)
        for k in values.keys():
            self.assertEqual(values[k], mammal_values[k],
                             "Object Import/Export disjoint")

        xml2 = "<folder>%s</folder>" % mammals_binding.renderXML(SET_ID)
        xml_list = xml.splitlines()
        xml2_list = xml2.splitlines()
        for x in xml2_list:
            self.assert_(x in xml_list, "Object Import Export disjoin")
        for x in xml_list:
            self.assert_(x in xml2_list, "Object Import Export disjoin")


class TestCataloging(MetadataTests):
    pass


class TestMetadataElement(MetadataTests):

    def testGetDefault(self):
        pm = getToolByName(self.root, 'portal_metadata')
        collection = pm.getCollection()
        set = collection.getMetadataSet(SET_ID)
        element = set.getElement('Title')
        element.field._edit_tales({'default':
                                   TALESMethod('content/getPhysicalPath')})
        zoo = self.root.zoo
        binding = pm.getMetadata(zoo)
        defaults = set.getDefaults(content = zoo)
        self.assertEqual(defaults['Title'],
                         zoo.getPhysicalPath(),
                         "Tales Context Passing Failed")

    def testGetDefaultWithTalesDelegate(self):
        mtool = getToolByName(self.root, 'portal_metadata')
        mtoolId = mtool.getId()
        collection = mtool.getCollection()
        set = collection.getMetadataSet(SET_ID)
        zoo = self.root.zoo
        test_value = 'Rabbits4Ever'
        binding = mtool.getMetadata(zoo)
        binding.setValues(SET_ID, {'Title':test_value})
        element = set.getElement('Description')
        # yikes, narly tales expression
        method = "python: content.%s.getMetadata(content).get(" \
                 "'%s', 'Title', no_defaults=1)" % (mtoolId, SET_ID)
        element.field._edit_tales({'default': TALESMethod(method)})
        value = binding.get(SET_ID, 'Description')
        self.assertEqual(value, test_value,
                         "Tales delegate for default didn't work")
        # make sure the right cached value was stored.
        value = binding.get(SET_ID, 'Description')
        self.assertEqual(value, test_value, "Wrong Data Object Cached")
        # test shortcut, too
        self.assertEqual(value,
                         mtool.getMetadataValue(zoo, SET_ID, 'Description'))

    def testAcquisitionInvariant(self):
        # invariant == can't be required and acquired
        from Products.SilvaMetadata.Exceptions import ConfigurationError
        pm = getToolByName(self.root, 'portal_metadata')
        collection = pm.getCollection()
        set = collection.getMetadataSet(SET_ID)
        element = set.getElement('Description')
        try:
            element.field._edit({'required':1})
            element.editElementPolicy(acquire_p = 1)
        except ConfigurationError:
            pass
        else:
            raise AssertionError("Acquisition/Required Element" \
                                 " Invariant Failed")
        try:
            element.field._edit({'required':0})
            element.editElementPolicy(acquire_p = 1)
            element.field._edit({'required':1})
        except ConfigurationError:
            pass
        else:
            raise AssertionError("Acquisition/Required Element" \
                                 " Invariant Failed 2")


class TestAdvancedMetadata(MetadataTests):
    """Tests for runtime binding methods"""

    def setupAcquiredMetadata(self):
        zoo = self.root.zoo
        binding = getToolByName(zoo, 'portal_metadata').getMetadata(zoo)
        set_id = SET_ID
        binding.setValues(set_id, {'Title':'hello world',
                                   'Description':'cruel place'})

    def testContainmentAcquisitionValue(self):
        self.setupAcquiredMetadata()
        zoo = self.root.zoo
        mams = zoo.mammals
        z_binding = getToolByName(zoo, 'portal_metadata').getMetadata(zoo)
        m_binding = getToolByName(mams, 'portal_metadata').getMetadata(mams)
        set_id = SET_ID
        self.assertEqual(m_binding[set_id]['Description'],
                         z_binding[set_id]['Description'])
        # test shortcut, too
        self.assertEqual(getToolByName(mams, 'portal_metadata').getMetadataValue(mams,set_id,'Description'),
                         getToolByName(zoo, 'portal_metadata').getMetadataValue(zoo,set_id,'Description'))
        self.assertNotEqual(m_binding.get(set_id, 'Description', acquire=0),
                            z_binding[set_id]['Description'])

    def testContainmentAcquisitionList(self):
        self.setupAcquiredMetadata()
        zoo = self.root.zoo
        mams = zoo.mammals
        m_binding = getToolByName(mams, 'portal_metadata').getMetadata(mams)
        z_binding = getToolByName(zoo, 'portal_metadata').getMetadata(zoo)
        set_id = SET_ID
        acquired = m_binding.listAcquired()
        # test the contained's list acquired
        self.assertEqual(len(acquired), 1)
        self.assertEqual(acquired[0][1], 'Description')
        # test the container's listacquired
        acquired = z_binding.listAcquired()
        self.assertEqual(len(acquired), 0)
        # special case for
        z_binding.setValues(set_id, {'Title':'', 'Description':''})
        acquired = z_binding.listAcquired()
        self.assertEqual(len(acquired), 0)

    def testObjectDelegation(self):
        from Acquisition import Implicit
        class Delegator(Implicit):
            def __init__(self, name):
                self.name = name
            def __call__(self):
                ob = self.aq_inner.aq_parent
                return getattr(ob, self.name)

        zoo = self.root.zoo
        delegate = Delegator('reptiles')
        zoo.delegate = delegate
        mams = zoo.mammals
        reps = zoo.reptiles
        r_binding = getToolByName(reps, 'portal_metadata').getMetadata(reps)
        m_binding = getToolByName(mams, 'portal_metadata').getMetadata(mams)
        r_binding.setValues(SET_ID,
                            {'Title':'snake food',
                             'Description':'yummy n the tummy'}
                            )
        m_binding.setObjectDelegator('delegate')
        self.assertEqual(
            m_binding[SET_ID]['Title'],
            r_binding[SET_ID]['Title']
            )
        # test shortcut, too
        self.assertEqual('snake food',
                         getToolByName(mams, 'portal_metadata').
                              getMetadataValue(mams, SET_ID, 'Title'))        
        m_binding.clearObjectDelegator()
        assert m_binding[SET_ID]['Title'] != r_binding[SET_ID]['Title']

    def testMutationTriggerDelegation(self):
        class MutationTrigger:
            def __init__(self):
                self.called = 0
            def __call__(self):
                self.called += 1

        zoo = self.root.zoo
        mams = zoo.mammals
        m_binding = getToolByName(mams, 'portal_metadata').getMetadata(mams)
        trigger = MutationTrigger()
        zoo.trigger = trigger
        m_binding.setMutationTrigger(SET_ID, 'Title', 'trigger')
        m_binding.setValues(SET_ID, {'Title':'surfin betty',
                                     'Description':'morning pizza'})
        self.assertEqual(trigger.called, 1)
        m_binding.setValues(SET_ID, {'Description':'midnight raid'})
        self.assertEqual(trigger.called, 1)
        m_binding.clearMutationTrigger(SET_ID)
        # props to tennyson
        m_binding.setValues(SET_ID,
                            {'Title':
                             'To strive, to seek, to find, and not to yield'}
                            )
        self.assertEqual(trigger.called, 1)


def test_suite():
    suite = TestSuite()
    suite.addTest(makeSuite(TestSetImportExport))
    suite.addTest(makeSuite(TestObjectImportExport))
    suite.addTest(makeSuite(TestCataloging))
    suite.addTest(makeSuite(TestMetadataElement))
    suite.addTest(makeSuite(TestAdvancedMetadata))
    return suite


if __name__ == '__main__':
    main()




"""
Access handlers are used to create a metadata binding
adapter for a particular piece of content. handlers
may be registered either for a specific content
type or as a default accessor to be used when no
more specific content type is found.

signature for a handler is::

  def access_handler(tool, content_type, content)

and should return a metadata binding adapter or
raise an exception.

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from Binding import MetadataBindAdapter
from Compatibility import getContentType
from Exceptions import BindingError

_default_accessor = None
_typeAccessHandlers = {}

def registerAccessHandler(content_type, handler):
    assert callable(handler)
    global _typeAccessHandlers, _default_accessor

    if content_type is None:
        _default_accessor = handler
    else:
        _typeAccessHandlers[content_type]=handler

def getAccessHandler(content_type):
    handler = _typeAccessHandlers.get(content_type)
    if handler is None:
        return _default_accessor
    return handler

def invokeAccessHandler(tool, content):
    ct = getContentType(content)
    handler = getAccessHandler(ct)
    if handler is None:
        raise BindingError("no access handler found for %s" % ct)
    return handler(tool, ct, content)

def default_accessor(tool, content_type, content):
    type_mapping = tool.getTypeMapping()
    metadata_sets = type_mapping.getMetadataSetsFor(content_type)

    if not metadata_sets:
        raise BindingError("no metadata sets defined for %s" % content_type)

    return MetadataBindAdapter(content, metadata_sets).__of__(content)


def cached_accessor(tool, content_type, content):
    """
    bindings that get cached
    """
    binding  = getattr(content, '_v_binding', None)

    if binding is None:
        binding = default_accessor( tool, content_type, content)
        setattr(content, '_v_binding', binding)
    return binding


registerAccessHandler(None, default_accessor)



"""
Common Zope Imports

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
from AccessControl import ClassSecurityInfo
from Acquisition import Implicit, aq_inner, aq_parent
from ComputedAttribute import ComputedAttribute
from OFS.SimpleItem import SimpleItem
from OFS.Folder import Folder
from Globals import DTMLFile, InitializeClass
from Interface import Base as Interface
from ZODB.PersistentMapping import PersistentMapping

from Compatibility import getToolByName, UniqueObject

# py2.2.2 forward decl
try:
    True, False
except:
    True = 1
    False = 0




"""
Simple Type System for (De)Serializing python values to XML

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

import types
from cgi import escape

from DateTime import DateTime
from Exceptions import XMLMarshallError

def serialize(value):

    if isinstance(value, types.IntType):
        return '<element type="integer">%s</element>' % unicode(str(value))

    elif isinstance(value, types.FloatType):
        return '<element type="float">%s</element>' % unicode(str(value))

    elif isinstance(value, types.StringType):
        return '<element type="string">%s</element>' % unicode(escape(value, 1))

    elif isinstance(value, DateTime):
        return '<element type="date">%s</element>' % unicode(str(value))

    elif isinstance(value, types.UnicodeType):
        return '<element type="string">%s</element>' % escape(value, 1)

    elif isinstance(value, types.ListType):
        return '<element_list type="list">%s</element_list>' \
               % (''.join(map(serialize, value)))

    elif isinstance(value, types.NoneType):
        return '<element type="none">%s</element>' % unicode(str(value))

def deserialize(node):
    if not node.nodeName in ('element', 'element_list'):
        raise XMLMarshallError("invalid xml node type %s" % node.nodeName)

    node.normalize() # Join adjacent text nodes
    childs = node.childNodes
    if len(childs):
        node_value = node.childNodes[0].nodeValue
    else:
        # There's no value set for this node
        return None

    node_type = node.getAttribute('type')

    try:
        if node_type == 'string':
            return node_value
        elif node_type == 'integer':
            return int(node_value)
        elif node_type == 'float':
            return float(node_value)
        elif node_type == 'date':
            return DateTime(node_value)
        elif node_type == 'list':
            res = []
            for cn in node.childNodes:
                res.append(
                    deserialize(cn)
                    )
            return res

    except Exception, e:
        raise XMLMarshallError(
            "error on marshalling %s %s, childnodes: %s (%s %s)" % (
                node.nodeName,
                node_type,
                len(node.childNodes),
                str(node),
                str(e)
                )
            )




"""
Provides for code compatiblity between silva and the cmf.

Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from Acquisition import aq_get
from OFS.ObjectManager import UNIQUE
from Interface import Interface
from ExtensionClass import Base

import Configuration
from Exceptions import CompatibilityException

#################################
### Interface declarations/assertions
if Configuration.UsingCMF:
    from Products.CMFCore.interfaces import portal_actions, portal_metadata
    IActionProvider = portal_actions.ActionProvider
    IPortalMetadata = portal_metadata.portal_metadata
else:
    class IActionProvider(Interface): pass
    class IPortalMetadata(Interface): pass

#################################
### Action Provider and Info
if Configuration.UsingCMF:
    from Products.CMFCore.ActionInformation import ActionInformation
    from Products.CMFCore.ActionProviderBase import ActionProviderBase
else:
    ActionInformation = None

    class ActionProviderBase(Base):
        pass

def actionFactory(**kw):
    if ActionInformation is None:
        return None
    return ActionInformation(**kw)

#################################
### Service Lookup

if Configuration.UsingCMF:
    from Products.CMFCore.utils import getToolByName

else:
    _marker = []

    SilvaToolMap = {
        'portal_catalog':'service_catalog',
        'portal_annotations':'service_annotations',
        'portal_metadata':'service_metadata'
        }

    def getToolByName(ctx, service_name, default=_marker):
        try:
            silva_name = SilvaToolMap[service_name]
        except KeyError, e:
            raise CompatibilityException(str(e))

        try:
            tool = aq_get(ctx, silva_name, default, 1)
        except AttributeError:
            if default is _marker:
                raise
            return default

        if tool is _marker:
            raise AttributeError(silva_name)

        return tool

#################################
### Content Type Lookup
if Configuration.UsingCMF:
    def getContentTypeNames(ctx):
        pt = getToolByName(ctx, 'portal_types')
        return pt.objectIds()
else:
    _allowed_content_types = []

    def registerTypeForMetadata(type_name):
        if type_name not in _allowed_content_types:
            _allowed_content_types.append(type_name)

    def getContentTypeNames(ctx):
        return tuple(_allowed_content_types)
        #return ctx.get_silva_addables_all()

if Configuration.UsingCMF:
    def getContentType(content):
        return content.getPortalTypeName()

else:
    def getContentType(content):
        return content.meta_type

#################################
### Permissions
# only set if not overridden by user/developer
if Configuration.UsingCMF:
    from Products.CMFCore import CMFCorePermissions
    if Configuration.pMetadataView is None:
        Configuration.pMetadataView = CMFCorePermissions.View
    if Configuration.pMetadataEdit is None:
        Configuration.pMetadataEdit = CMFCorePermissions.ModifyPortalContent
    if Configuration.pMetadataManage is None:
        Configuration.pMetadataManage = CMFCorePermissions.ManagePortal
else:
    from Products.Silva import SilvaPermissions
    if Configuration.pMetadataView is None:
        Configuration.pMetadataView = SilvaPermissions.View
    if Configuration.pMetadataEdit is None:
        Configuration.pMetadataEdit = SilvaPermissions.ChangeSilvaContent
    if Configuration.pMetadataManage is None:
        Configuration.pMetadataManage = SilvaPermissions.ViewManagementScreens

#################################
### Catalog Expressions for ProxyIndex
if Configuration.UsingCMF:
    tem = "python: object.portal_metadata.getMetdatata(object)['%s']['%s']"
    index_expression_template = tem
else:
    tem = "python: object.service_metadata.getMetadataValue(object, '%s', '%s')"
    index_expression_template = tem

#################################
### Misc
if Configuration.UsingCMF:
    from Products.CMFCore.utils import UniqueObject
else:
    class ImmutableId(Base):

        """ Base class for objects which cannot be renamed.
        """
        def _setId(self, id):

            """ Never allow renaming!
            """
            if id != self.getId():
                raise MessageDialog(
                    title='Invalid Id',
                    message='Cannot change the id of this object',
                    action ='./manage_main',)

    class UniqueObject (ImmutableId):

        """ Base class for objects which cannot be "overridden" / shadowed.
        """
        __replaceable__ = UNIQUE




MetadataNamespace = 'http://www.infrae.com/xml/metadata'
BindingRunTime = 'http://www.infrae.com/xml/metadata/binding'
DefaultPrefix = 'example'
DefaultNamespace = "http://www.example.com/unknown_namespace"
DublinCore = 'http://www.dublincore.org'




"""
Type Specific Initialization of Metadata Binding Runtime Data
not really meant for public api exposure (certainly not ttw)

Handler Signature is as follows

def metadata_initialization(content, bind_data)

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from Compatibility import getContentType

_default_initializer = None

_typeInitHandlers = {}

def registerInitHandler(content_type, handler):
    assert callable(handler)
    global _typeInitHandlers, _default_initializer

    if content_type is None:
        _default_initializer = handler
    else:
        _typeInitHandlers[content_type]=handler

def getHandler(content):
    ct = getContentType(content)

    handler = _typeInitHandlers.get(ct)

    if handler is None and _default_initializer:
        return _default_initializer

    # possibly none
    return handler





"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from Acquisition import aq_inner, aq_parent
from AccessControl import getSecurityManager, ClassSecurityInfo
from Globals import InitializeClass
from Persistence import Persistent
from Products.PageTemplates.Expressions import getEngine
from Products.PageTemplates.Expressions import SecureModuleImporter


class Expression (Persistent):
    text = ''
    _v_compiled = None

    security = ClassSecurityInfo()

    def __init__(self, text):
        self.text = text
        self._v_compiled = getEngine().compile(text)

    def __call__(self, econtext):
        compiled = self._v_compiled
        if compiled is None:
            compiled = self._v_compiled = getEngine().compile(self.text)
        res = compiled(econtext)
        if isinstance(res, Exception):
            raise res
        return res

InitializeClass(Expression)

def createExprContext(element, ob):
    '''
    An expression context provides names for TALES expressions.
    '''

    data = {
        'here':         ob,
        'container':    aq_parent(aq_inner(ob)),
        'nothing':      None,
        'element':      element,
        'request':      getattr(ob, 'REQUEST', None),
        'modules':      SecureModuleImporter,
        'user':         getSecurityManager().getUser(),
        }
    return getEngine().getContext(data)









"""
Metadata Elements
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from AccessControl import getSecurityManager
import Configuration
from Exceptions import ConfigurationError
from FormulatorField import getFieldFactory
from Guard import Guard
from Interfaces import IMetadataElement
from ZopeImports import *
from utils import normalize_kv_pairs

_marker = []

encoding = 'UTF-8'


class MetadataElement(SimpleItem):
    """
    Property Bag For Element Policies, this implementation is
    formulator specific
    """

    meta_type = 'Metadata Element'

    __implements__ = IMetadataElement

    #################################
    # default element policy properties
    #################################

    read_only_p = False
    index_p = False
    acquire_p = False
    index_type = None
    field_type = None
    field = None

    ## defer to formulator for now
    #use_default_p = True
    #required_p = False
    #default = None

    ## out of scope for initial impl
    #export_p = True
    #enforce_vocabulary_p = True

    manage_options = (
        {'label':'Settings',
         'action':'manage_settings'},

        {'label':'Guards',
         'action':'manage_guard_form'},

        {'label':'Field',
         'action':'field/manage_main'},
        )

    security = ClassSecurityInfo()

    security.declareProtected(Configuration.pMetadataManage, 'manage_settings')
    manage_settings = DTMLFile('ui/ElementPolicyForm', globals())

    security.declareProtected(Configuration.pMetadataManage,
                              'manage_guard_form')
    manage_guard_form = DTMLFile('ui/ElementGuardForm', globals())


    def __init__(self, id, **kw):
        self.id = id
        self.read_guard = Guard()
        self.write_guard = Guard()
        self.index_constructor_args = {}

        self.editElementPolicy(**kw)

    def editElementGuards(self, read_guard, write_guard, RESPONSE=None):

        self.read_guard.changeFromProperties(read_guard)
        self.write_guard.changeFromProperties(write_guard)

        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    def editElementPolicy(self,
                          field_type = None,
                          index_type = None,
                          index_p = None,
                          read_only_p = None,
                          extra = None,
                          acquire_p = None,
                          RESPONSE = None
                          ):
        """
        edit an element's policy
        """
        if index_type is not None:
            ms = self.getMetadataSet()
            if ms.isInitialized():
                raise ConfigurationError("Not Allowed Set Already initialized")

        if field_type is not None:
            ms = self.getMetadataSet()
            if ms.isInitialized():
                raise ConfigurationError("Not Allowed Set Already initialized")

        f = self.field
        if acquire_p is not None and f is not None:
            required_p = f.has_value('required') and f.get_value('required')
            if required_p and acquire_p:
                raise ConfigurationError("Required Values may not be Acquired")

        field_type = field_type or self.field_type
        index_type = index_type or self.index_type

        if index_p is None:
            index_p = self.index_p

        if read_only_p is None:
            read_only_p = self.read_only_p

        if acquire_p is None:
            acquire_p = self.acquire_p

        if field_type != self.field_type:
            try:
                factory = getFieldFactory(field_type)
            except KeyError:
                raise ConfigurationError("invalid field type %s" % field_type)
            self.field = factory(self.getId())
            self.field.field_record = self.getMetadataSet().getId()
            self.field_type = field_type
            self.field.values['unicode']=1

        if index_type is not None:
            if index_type in self.getMetadataSet().listIndexTypes():
                self.index_type = index_type
            else:
                raise ConfigurationError("invalid index type %s" % index_type)

        # need to cascacde this so we can create indexes at the set level
        self.index_p = not not index_p
        self.read_only_p = not not read_only_p
        self.acquire_p = not not acquire_p

        # normalize the key value pairs into a mapping[k]->v
        if extra:
            constructor_args = normalize_kv_pairs(extra)
            self.index_constructor_args.update(constructor_args)

        if RESPONSE is not None:
            return RESPONSE.redirect('manage_workspace')

    def validate(self, data):
        return self.field.validate(data)

    def Title(self):
        return self.field.get_value('title')

    def Description(self):
        return self.field.get_value('description')

    def isViewable(self, content):
        """
        is this element viewable for the content object
        """
        return self.read_guard.check(getSecurityManager(), self, content)

    def isEditable(self, content):
        """
        is this element editable for the content object
        """
        if not self.read_only_p:
            return self.write_guard.check(getSecurityManager(), self, content)

    def isAcquireable(self):
        """
        is this field acquireable to objects lower in the container
        """
        return self.acquire_p

    def renderView(self, value=None):
        """
        render the element given a particular element value
        """
        return self.field.render_view(value)

    def renderEdit(self, value=None):
        """
        render the element as a form field given a particular value
        """
        return self.field.render(value)

    def isRequired(self):
        """
        is the element required to have a value
        """
        return self.field.is_required()

    def getDefault(self, content):
        """
        return the default value for this element
        """
        return self.field.get_value('default', content=content)

    def getVocabulary(self, content):
        """
        return the set of allowed of words
        """
        raise NotImplemented

    ## little hack to get formulator fields to do unicode
    def get_form_encoding(self):
        return encoding

    # fields are not in unicode mode, unfortunately
    def get_unicode_mode(self):
        return 0
    
    ## formulator sends on change messages when internal field
    ## values are changed, we catch the required message in
    ## order to maintain our invariant that an element can not
    ## be both required and acquired.
    def on_value_required_changed(self, required_value):
        if required_value and self.acquire_p:
            raise ConfigurationError("Acquired values may not be Required")

InitializeClass(MetadataElement)

def ElementFactory(id, **kw):

    return MetadataElement(id, **kw)



"""
provides access to formulator field registry

author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
from Products.Formulator.FieldRegistry import FieldRegistry

def getFieldFactory(fieldname):
    return FieldRegistry.get_field_class(fieldname)

def listFields():
    mapping = FieldRegistry.get_field_classes()
    field_types = mapping.keys()
    field_types.sort()
    return field_types





"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from ZopeImports import *
from Exceptions import NotFound
from Interfaces import IMetadataCollection
from Import import read_set, make_set
from Set import MetadataSet
from Configuration import pMetadataManage

class MetadataCollection(Folder):

    meta_type = 'Metadata Collection'

    __implements__ = IMetadataCollection

    security = ClassSecurityInfo()

    all_meta_types = (
        {'name':MetadataSet.meta_type,
         'action':'addMetadataSetForm'},
        )

    manage_options = (

        {'label':'Metadata Sets',
         'action':'manage_main'},

        {'label':'Metadata Tool',
         'action':'../manage_workspace'},
        )

    security.declareProtected(pMetadataManage, 'addMetadataSetForm')
    addMetadataSetForm = DTMLFile('ui/MetadataSetAddForm', globals())

    security.declareProtected(pMetadataManage, 'addMetadataSet')
    def addMetadataSet(self, id, namespace_prefix, namespace_uri,
                       title='', description='', RESPONSE=None):
        " "

        set = MetadataSet(id=id,
                          title = title,
                          description = description,
                          metadata_prefix = namespace_prefix,
                          metadata_uri = namespace_uri,
                          )

        self._setObject(id, set)

        if RESPONSE is not None:
            return self.manage_main(update_menu=1)

    #security.declareProtected()
    def getMetadataSets(self):
        return self.objectValues('Metadata Set')

    #security.declareProtected()
    def getMetadataSet(self, set_id):
        return self._getOb(set_id)

    #security.declareProtected()
    def getSetByNamespace(self, metadata_uri):
        for set in self.getMetadataSets():
            if set.metadata_uri == metadata_uri:
                return set

        raise NotFound("No Metadata Set Matching %s" % str(metadata_uri))

    security.declareProtected(pMetadataManage, 'importSet')
    def importSet(self, xml_file, RESPONSE=None):
        """ import an xml definition of a metadata set"""

        set_node = read_set(xml_file)
        make_set(self, set_node)

        if RESPONSE is not None:
            return self.manage_main(update_menu=1)


InitializeClass(MetadataCollection)



from Products.SilvaMetadata.MetadataTool import MetadataTool
from Products.SilvaMetadata.utils import StringBuffer

def install(self):

    out = StringBuffer()

    print >> out, "Installing Silva Metadata Service"

    silva_id = 'service_metadata'
    root = self.get_root()
    ob = MetadataTool()
    ob.id = silva_id
    root._setObject(silva_id, ob)

    print >> out, "Installation Complete"
    return out.getvalue()




"""
Marker Interfaces
author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
from Interface import Interface
from Compatibility import IPortalMetadata

#################################
# Metadata Tool/Service Interface
#################################
class IMetadataTool(IPortalMetadata):
    pass

#################################
# Base Building Blocks
#################################

class IMetadataCollection(Interface):
    pass

class IOrderedContainer(Interface):

    def moveObject(id, position):
        """
        move an object with the given an id to the specified
        position.
        """
    def moveObjectUp(id, steps=1):
        """
        move an object with the given id up the ordered list
        the given number of steps
        """

    def moveObjectDown(id, steps=1):
        """
        move an object with the given id down the ordered list
        the given number of steps
        """

    def getObjectPosition(id):
        """
        given an object id return its position in the ordered list
        """

class IMetadataSet(Interface):
    pass

class IMetadataElement(Interface):
    pass

#################################
# Adapter Provided Functionality
#################################
# all of these operate on a set basis.

class IMetadataSetExporter(Interface):
    pass

class IMetadataForm(Interface):
    pass

class IMetadataValidation(Interface):
    pass

class IMetadataStorage(Interface):
    pass



"""
author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
import sys

from AccessControl import getSecurityManager, Permissions

from Compatibility import actionFactory
import Configuration
from Element import MetadataElement, ElementFactory
from Exceptions import NamespaceConflict, ConfigurationError
from Export import MetadataSetExporter
from FormulatorField import listFields
from Index import createIndexes
from Interfaces import IMetadataSet, IOrderedContainer
from Namespace import DefaultNamespace, DefaultPrefix
from ZopeImports import *

from Products.ProxyIndex.ProxyIndex import getIndexTypes

class OrderedContainer(Folder):

    __implements__ = IOrderedContainer

    security = ClassSecurityInfo()

    security.declareProtected(Permissions.copy_or_move, 'moveObject')
    def moveObject(self, id, position):
        obj_idx  = self.getObjectPosition(id)
        if obj_idx == position:
            return None
        elif position < 0:
            position = 0

        metadata = list(self._objects)
        obj_meta = metadata.pop(obj_idx)
        metadata.insert(position, obj_meta)
        self._objects = tuple(metadata)

    security.declareProtected(Permissions.copy_or_move, 'getObjectPosition')
    def getObjectPosition(self, id):

        objs = list(self._objects)
        om = [objs.index(om) for om in objs if om['id']==id ]

        if om: # only 1 in list if any
            return om[0]

        raise NotFound('Object %s was not found' % str(id))

    security.declareProtected(Permissions.copy_or_move, 'moveObjectUp')
    def moveObjectUp(self, id, steps=1, RESPONSE=None):
        """ Move an object up """
        self.moveObject(
            id,
            self.getObjectPosition(id) - int(steps)
            )
        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    security.declareProtected(Permissions.copy_or_move, 'moveObjectDown')
    def moveObjectDown(self, id, steps=1, RESPONSE=None):
        """ move an object down """
        self.moveObject(
            id,
            self.getObjectPosition(id) + int(steps)
            )
        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    security.declareProtected(Permissions.copy_or_move, 'moveObjectTop')
    def moveObjectTop(self, id, RESPONSE=None):
        """ move an object to the top """
        self.moveObject(id, 0)
        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    security.declareProtected(Permissions.copy_or_move, 'moveObjectBottom')
    def moveObjectBottom(self, id, RESPONSE=None):
        """ move an object to the bottom """
        self.moveObject(id, sys.maxint)
        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    def manage_renameObject(self, id, new_id, REQUEST=None):
        " "
        objidx = self.getObjectPosition(id)
        method = OrderedContainer.inheritedAttribute('manage_renameObject')
        result = method(self, id, new_id, REQUEST)
        self.moveObject(new_id, objidx)

        return result

InitializeClass(OrderedContainer)

class MetadataSet(OrderedContainer):
    """
    Set of Elements constituting a metadata dialect
    """

    meta_type = 'Metadata Set'
    __implements__ = IMetadataSet

    security = ClassSecurityInfo()

    all_meta_types = (
        {'name':MetadataElement.meta_type,
         'action':'addElementForm'},
        )

    manage_options = (
        {'label':'Elements',
         'action':'manage_main'},
        {'label':'Settings',
         'action':'manage_settings'},
        {'label':'Action',
         'action':'manage_action'},
        )

    security.declareProtected(Configuration.pMetadataManage, 'manage_settings')
    manage_settings = DTMLFile('ui/SetSettingsForm', globals())

    security.declareProtected(Configuration.pMetadataManage, 'manage_action')
    manage_action  = DTMLFile('ui/SetActionForm', globals())

    security.declareProtected(Configuration.pMetadataManage, 'addElementForm')
    addElementForm  = DTMLFile('ui/ElementAddForm', globals())

    manage_main = DTMLFile('ui/SetContainerView', globals())

    initialized = None
    use_action_p = None
    action = None
    title = ''
    description = ''
    
    # for backwards compatibility...
    _category = ''
    _minimal_role = ''

    def __init__(self, id, title='', description='',
                 metadata_prefix=DefaultPrefix, metadata_uri=DefaultNamespace):

        self.id = id
        self.initialized = None
        self.use_action_p = None
        self.title = ''
        self.description = ''
        self._minimal_role = ''
        self._category = ''
        
        # we can't do any verification till after we have a ctx
        self.metadata_uri = metadata_uri
        self.metadata_prefix = metadata_prefix


    def getTitle(self):
        return self.title

    def getDescription(self):
        return self.description

    def getMinimalRole(self):
        return self._minimal_role

    def setMinimalRole(self, role):
        self._minimal_role = role

    def getCategory(self):
        return self._category

    def setCategory(self, cat):
        self._category = cat
        
    def addMetadataElement(self,
                           id,
                           field_type,
                           index_type,
                           index_p=None,
                           acquire_p=None,
                           read_only_p=None,
                           RESPONSE=None):
        """ """
        element = ElementFactory(id)
        self._setObject(id, element)
        element = self._getOb(id)

        element.editElementPolicy(field_type = field_type,
                                  index_type = index_type,
                                  index_p = index_p,
                                  read_only_p = read_only_p,
                                  acquire_p = acquire_p)

        if RESPONSE is not None:
            return RESPONSE.redirect('manage_main')

    def editAction(self,
                   use_action_p,
                   id,
                   permission,
                   action,
                   condition,
                   category,
                   visible,
                   RESPONSE=None):
        """ CMF Action Provider Support """

        self.use_action_p = not not use_action_p

        self.action = actionFactory(id=id,
                                    title=title,
                                    permission=permission,
                                    category=category,
                                    action=action,
                                    condition=condition,
                                    visible=visible)

        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    def editSettings(
        self, title, description, ns_uri, ns_prefix, 
        minimal_role='', category=''):
        """ Edit Set Settings """

        if self.isInitialized():
            raise ConfigurationError (" Set Already Initialized ")

        self.title = title
        self.description = description
            
        self.setNamespace(ns_uri, ns_prefix)
        self.setMinimalRole(minimal_role)
        self.setCategory(category)
            
        request = getattr(self, 'REQUEST', None)
        if request is not None:
            request.RESPONSE.redirect('manage_workspace')

    def exportXML(self, RESPONSE=None):
        """ export an xml serialized version of the policy """

        exporter = MetadataSetExporter(self)

        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/xml')
            RESPONSE.setHeader('Content-Disposition',
                               'attachment; filename=%s.xml' % self.getId())
        return exporter()

    def setNamespace(self, ns_uri, ns_prefix):
        verifyNamespace(self, ns_uri, ns_prefix)
        self.metadata_prefix = ns_prefix
        self.metadata_uri = ns_uri

    def isInitialized(self):
        return self.initialized

    def setInitialized(self, initialization_flag, RESPONSE=None):
        """ """
        flag = not not initialization_flag

        if flag != self.initialized:
            if self.initialized:
                self.initialized = 0
            else:
                self.initialize()

        if RESPONSE:
            return RESPONSE.redirect('manage_workspace')

    def initialize(self, RESPONSE=None):
        """ initialize the metadata set """
        if self.isInitialized():
            return None

        # install indexes
        indexables = [e for e in self.getElements() if e.index_p]
        catalog = getToolByName(self, 'portal_catalog')
        createIndexes(catalog, indexables)

        self.initialized = 1

        if RESPONSE is not None:
            RESPONSE.redirect('manage_workspace')

    def getNamespace(self):
        return (self.metadata_prefix, self.metadata_uri)

    def getElements(self):
        return self.objectValues(spec='Metadata Element')

    def getElement(self, element_id):
        return self._getOb(element_id)

    def getElementsFor(self, object, mode='view'):

        if mode == 'view':
            guard = 'read_guard'
        else:
            guard = 'write_guard'

        sm = getSecurityManager()

        res = []
        for e in self.getElements():
            if getattr(e, guard).check(sm, e, object):
                res.append(e)
            if mode == 'edit' and e.read_only_p:
                continue
        return res

    def getMetadataSet(self):
        return self

    def getDefaults(self, content):
        res = {}
        for e in self.getElements():
            res[e.getId()] = e.getDefault(content)
        return res

    def listFieldTypes(self):
        return listFields()

    def listIndexTypes(self):
        return getIndexTypes(getToolByName(self, 'portal_catalog'))

    def manage_afterAdd(self, item, container):
        # verify our namespace
        self.setNamespace(self.metadata_uri, self.metadata_prefix)

InitializeClass(MetadataSet)

def verifyNamespace(ctx, uri, prefix):

    sid = ctx.getId()
    container = aq_parent(aq_inner(ctx))

    for s in container.getMetadataSets():
        if s.getId() == sid:
            continue
        if s.metadata_uri == uri:
            raise NamespaceConflict("%s uri is already in use" % uri)
        elif s.metadata_prefix == prefix:
            raise NamespaceConflict("%s prefix is already in use" % prefix)



"""
author: kapil thangavelu <k_vertigo@objectrealms.net>
"""

from xml.sax import make_parser, ContentHandler
from xml.dom import XMLNS_NAMESPACE
from UserDict import UserDict
from XMLType import deserialize
from Exceptions import NotFound

_marker = []

class DefinitionNode(UserDict):
    reserved = ('data',)

    def __getattr__(self, name):

        v = self.__dict__.get(name, _marker)
        if v is _marker:
            return self.data[name]
        return v

    def __setattr__(self, name, value):
        if name in DefinitionNode.reserved:
            self.__dict__[name]=value
        else:
            self.data[name]=value


class MetaReader(ContentHandler):

    def __init__(self):

        self.buf = []
        self.prefix = ''
        self.stack = []

    def startElement(self, element_name, attrs):
        name = element_name.lower()
        if self.prefix:
            name = '%s%s' % (self.prefix, name.capitalize())
        else: 
            name = name.capitalize()

        method = getattr(self, 'start%s' % name, None)
        # get rid of unicode...
        d = {}
        for k, v in attrs.items():
            d[str(k)]= str(v)

        if method:
            apply(method, (d,))

    def endElement(self, element_name):
        chars = str(''.join(self.buf)).strip()
        self.buf = []
        name = element_name.lower()
        if self.prefix: 
            name = '%s%s' % (self.prefix, name.capitalize())
        else:
            name = name.capitalize()
        
        method = getattr(self, 'end%s' % name, None)
        if method:
            apply(method, (chars,))

    def characters(self, chars):
        self.buf.append(chars)


class MetadataSetReader(MetaReader):

    def getSet(self):
        return self.set

    def endTitle(self, chars):
        self.set.title = chars

    def endDescription(self, chars):
        self.set.description = chars

    def endMinimalrole(self, chars):
        self.set.minimalrole = chars
        
    def endCategory(self, chars):
        self.set.category = chars
        
    def startMetadata_set(self, attrs):
        self.set = s = DefinitionNode(attrs)
        s.setdefault('elements', [])

    def startMetadata_element(self, attrs):
        element = DefinitionNode(attrs)
        self.set.elements.append(element)

    def getElement(self):
        return self.set.elements[-1]

    def endIndex_type(self, chars):
        self.getElement().index_type = chars

    def endIndex_p(self, chars):
        self.getElement().index_p = chars

    def endRead_only_p(self, chars):
        self.getElement().read_only_p = int(chars)

    def endAcquire_p(self, chars):
        self.getElement().acquire_p = int(chars)

    def endField_type(self, chars):
        self.getElement().field_type = chars

    def startRead_guard(self, attrs):
        self.prefix = 'Read'
        self.getElement().read_guard = DefinitionNode()

    def endReadRead_guard(self, chars):
        self.prefix = ''

    def endReadRoles(self, chars):
        self.getElement().read_guard.roles = chars

    def endReadPermissions(self, chars):
        self.getElement().read_guard.permissions = chars

    def endReadExpr(self, chars):
        self.getElement().read_guard.expr = chars

    def startWrite_guard(self, attrs):
        self.getElement().write_guard = DefinitionNode()
        self.prefix = 'Write'

    def endWriteWrite_guard(self, chars):
        self.prefix = ''

    def endWriteRoles(self, chars):
        self.getElement().read_guard.roles = chars

    def endWritePermissions(self, chars):
        self.getElement().read_guard.permissions = chars

    def endWriteExpr(self, chars):
        self.getElement().read_guard.expr = chars

    def startField_values(self, attrs):
        self.getElement().field_values = []
        self.prefix = 'FieldV'

    def endFieldVField_values(self, chars):
        self.prefix = ''

    def startFieldVValue(self, attrs):
        fv = DefinitionNode(attrs)
        self.getElement().field_values.append(fv)

    def startField_tales(self, attrs):
        self.getElement().field_tales = []
        self.prefix = 'FieldT'

    def endFieldTField_tales(self, chars):
        self.prefix = ''

    def startFieldTValue(self, attrs):
        ft = DefinitionNode(attrs)
        self.getElement().field_tales.append(ft)

    def endFieldTValue(self, chars):
        self.getElement().field_tales[-1].expr = chars

    def startField_messages(self, attrs):
        self.getElement().field_messages = []
        self.prefix='FieldM'

    def endFieldMField_messages(self, chars):
        self.prefix = ''

    def startFieldMmessage(self, attrs):
        fm = DefinitionNode(attrs)
        self.getElement().field_messages.append(fm)

    def endFieldMmessage(self, chars):
        self.getElement().field_messages[-1].text = chars

    def startIndex_args(self, attrs):
        self.prefix = 'IndexArg'
        self.getElement().index_args = []

    def endIndexArgIndex_args(self, chars):
        self.prefix = ''

    def startIndexArgValue(self, attrs):
        iav = DefinitionNode(attrs)
        self.getElement().index_args.append(iav)

    def endIndexArgValue(self, chars):
        self.getElement().index_args[-1].value = chars


def read_set(xml):
    parser = make_parser()
    reader = MetadataSetReader()
    parser.setContentHandler(reader)
    parser.parse(xml)
    return reader.getSet()


def make_set(container, set_node):

    import Configuration
    from Compatibility import getToolByName
    from Products.Formulator.TALESField import TALESMethod

    # compatiblity.. ick
    if not set_node.has_key('title'):
        set_node['title'] = ''
    if not set_node.has_key('description'):
        set_node['description'] = ''

    pm = getToolByName(container, 'portal_metadata')

    collection = getattr(pm, Configuration.MetadataCollection)
    collection.addMetadataSet(
        set_node.id, set_node.ns_prefix, set_node.ns_uri, set_node.title,
        set_node.description)

    set = pm.getMetadataSet(set_node.id)
    set.setCategory(set_node.get('category', ''))
    set.setMinimalRole(set_node.get('minimalrole', ''))

    for e_node in set_node.elements:

        # compatiblity.. ick
        if not e_node.has_key('acquire_p'):
            e_node['acquire_p'] = 0
        if not e_node.has_key('read_only_p'):
            e_node['read_only_p'] = 0
        if not e_node.has_key('index_p'):
            e_node['index_p'] = 0
            
        # type possible is string, convert to 'boolean'
        for p in ['index_p', 'acquire_p', 'read_only_p']:
            try:
                e_node[p] = not not int(e_node[p])
            except ValueError:
                e_node[p] = 0

        set.addMetadataElement(
            e_node.id, e_node.field_type, e_node.index_type, e_node.index_p,
            e_node.acquire_p, e_node.read_only_p)

        element = set.getElement(e_node.id)

        field = element.field
        for fv in e_node.field_values:
            k = fv.key
            v = fv.value
            t = fv.type
            if t == 'int':
                v = int(v)
            elif t == 'float':
                v = float(v)
            elif t == 'list':
                # XXX this is incomplete support for lists
                # the originial version of formulator xml support
                # did an eval here, which is not an option
                # XXX we'll make use of this option anyway
                # metadata sets are edited on the filesystem after all
                v = eval(v, {})

            field.values[k]=v

        # some sort of formulator hack
        if e_node.field_type == 'DateTimeField':
            field.on_value_input_style_changed(field.get_value('input_style'))

        for ft in e_node.field_tales:
            if ft.expr:
                field.tales[ft.key] = TALESMethod(ft.expr)

        for fm in e_node.field_messages:
            field.message_values[fm.name]=fm.text

        constructor_args = {}
        for iav in e_node.index_args:
            constructor_args[iav['key']]=iav['value']
        element.index_constructor_args = constructor_args


def import_metadata(content, content_node):
    """

    minimal import system for object metadata

    """

    from Compatibility import getToolByName

    metadata_node = metadata_node_search(content_node)
    metadata_tool = getToolByName(content, 'portal_metadata')

    if not metadata_node:
        return

    metadata = {}
    binding = metadata_tool.getMetadata(content)

    for namespace_attr in metadata_node.attributes.values():
        if not namespace_attr.namespaceURI == XMLNS_NAMESPACE:
            continue
        metadata[namespace_attr.value] = {}

    for child in metadata_node.childNodes:
        if not child.namespaceURI:
            continue
        metadata[child.namespaceURI][child.localName] = deserialize(
            child.firstChild)

    for k, v in metadata.items():
        # First delete 'read-only' elements from dictionary
        try:
            set_name = binding.getSetNameByURI(k)
        except NotFound:
            continue
        element_names = v.keys()
        for element_name in element_names:
            if not binding.isEditable(set_name, element_name):
                del v[element_name]

        # Set data
        errors = binding._setData(namespace_key=k,
                                  data=v,
                                  reindex=1)

        errors = None # discard errors for now
        if errors:
            raise ValidationError(
                "%s %s" % (
                    str(content.getPhysicalPath()),
                    str(errors)
                    )
                )

def metadata_node_search(content_node):

    for child_node in content_node.childNodes:
        if child_node.nodeName == 'metadata':
            return child_node

    return None


if __name__ == '__main__':
    # visual check
    import sys, pprint
    set_node = read_set(sys.argv[1])

    for k,v in set_node.items():
        print k

        if isinstance(v, DefinitionNode):
            for k,v in v.items():
                print "  "*5, k, v
        else:
            print v



"""
Author: kapil thangavelu <k_vertigo@objectrealms.net>
"""
from cgi import escape
from types import IntType, FloatType, ListType

from Interfaces import IMetadataSetExporter

from Element import MetadataElement
from XMLType import serialize
from utils import StringBuffer, make_lookup

class MetadataSetExporter:
    """
    for exporting a metadata set definition
    """
    __implements__ = IMetadataSetExporter

    def __init__(self, set):
        self.set = set

    def __call__(self, out=None):

        ext_out = 1
        if out is None:
            ext_out = 0
            out = StringBuffer()

        print >> out, '<?xml version="1.0"?>\n\n'
        print >> out, '<metadata_set id="%s" ns_uri="%s" ns_prefix="%s">' % (
            self.set.getId(),
            self.set.metadata_uri,
            self.set.metadata_prefix
            )

        title = escape(self.set.getTitle(), 1)
        description = escape(self.set.getDescription(), 1)
        minimalrole = escape(self.set.getMinimalRole(), 1)
        category = escape(self.set.getCategory(), 1)
            
        out.write('<title>%s</title>\n' % (title))
        out.write('<category>%s</category>\n' % (category))
        out.write('<description>%s</description>\n' % (description))
        out.write('<minimalrole>%s</minimalrole>\n' % (minimalrole))
                   
        for e in self.set.getElements():

            print >> out, '  <metadata_element id="%s">' % e.getId()
            print >> out, '   <index_type>%s</index_type>' % e.index_type
            print >> out, '   <index_p>%d</index_p>' % e.index_p
            print >> out, '   <field_type>%s</field_type>' % e.field_type
            print >> out, '   <acquire_p>%d</acquire_p>' % e.acquire_p
            print >> out, '   <read_only_p>%d</read_only_p>' % e.read_only_p

            g = e.read_guard

            print >> out, '   <read_guard>'
            print >> out, '     <roles>%s</roles>' % g.getRolesText()
            print >> out, '     <permissions>%s</permissions>' \
                          % g.getPermissionsText()
            print >> out, '     <expr>%s</expr>' % g.getExprText()
            print >> out, '   </read_guard>'

            g = e.write_guard
            print >> out, '   <write_guard>'
            print >> out, '     <roles>%s</roles>' % g.getRolesText()
            print >> out, '     <permissions>%s</permissions>' \
                          % g.getPermissionsText()
            print >> out, '     <expr>%s</expr>' % g.getExprText()
            print >> out, '   </write_guard>'

            f = e.field

            print >> out, '   <field_values>'

            for k, v in f.values.items():
                if v is None:
                    continue
                if isinstance(v, IntType):
                    out.write('        ' \
                              '<value key="%s" type="%s" value="%d" />\n'
                               % (k, 'int', v))
                elif isinstance(v, FloatType):
                    out.write('        ' \
                              '<value key="%s" type="%s" value="%d" />\n'
                              % (k, 'float', v))
                elif isinstance(v, ListType):
                    out.write('        ' \
                              '<value key="%s" type="%s" value="%s" />\n'
                              % (k, 'list', escape(str(v), 1)))
                else:
                    out.write('        ' \
                              '<value key="%s" type="str" value="%s" />\n'
                              % (k, escape(str(v), 1)))

            print >> out, '   </field_values>'

            print >> out, '   <field_tales>'
            for k,v in f.tales.items():
                if v is None:
                    continue
                # FIXME: we get to the actual "source" for the TALESMethod by
                # getting its _text
                # Needs a different way of retrieving this value?
                out.write('     <value key="%s">%s</value>\n'
                          % (k, escape(getattr(v, '_text', ''), 1)))
            print >> out, '   </field_tales>'

            print >> out, '   <field_messages>'
            for message_key in f.get_error_names():
                message_text = f.get_error_message(
                    message_key, want_message_id=False)
                out.write('     <message name="%s">%s</message>\n' % (
                    escape(message_key, 1), escape(message_text, 1)))

            print >> out, '   </field_messages>'

            print >> out, '   <index_args>'
            for k,v in e.index_constructor_args.items():
                out.write('     <value key="%s">%s</value>\n'
                          % (k, escape(str(v), 1)))
            print >> out, '   </index_args>'

            print >> out, '  </metadata_element>'



        print >> out, '</metadata_set>'

        if ext_out:
            return out

        return out.getvalue()

class ObjectMetadataExporter:
    """
    for exporting the metadata of an object, returns
    an xml fragement.

    XXX encoding issues
    XXX file/image binary metadata issues
    """

    def __init__(self, binding, sets):
        self.binding = binding
        self.sets = sets

    def __call__(self, out=None):

        external_out = not not out

        if out is None:
            out = StringBuffer()

        out.write('<metadata \n')

        for set in self.sets:
            out.write('    xmlns:%s="%s"\n'
                      % (set.metadata_prefix, set.metadata_uri))
        out.write('      >')

        for set in self.sets:
            sid = set.getId()
            prefix = set.metadata_prefix
            check = make_lookup(set.objectIds(MetadataElement.meta_type))
            data = self.binding._getData(sid)

            for k,v in data.items():
                # with updates its possible that certain annotation data
                # is no longer part of the metadata set, so we filter it out.
                if not check(k):
                    continue

                out.write(u'      <%s:%s>%s</%s:%s>\n'
                          % (prefix, k, serialize(v), prefix, k))

        out.write('</metadata>\n')

        if external_out:
            return None

        return out.getvalue()
